from argparse import ArgumentParser
from io import BytesIO, StringIO
import json
from pathlib import Path
import tarfile
from urllib.error import HTTPError
from urllib.request import urlopen
from wheel.wheelfile import WheelFile
from zipfile import ZipFile, ZipInfo, ZIP_DEFLATED

METADATA = {
    "Summary": "Austin - Frame Stack Sampler for CPython",
    "Author": "Gabriele N. Tornetta",
    "License": "GPLv3+",
    "Classifier": [
        "Development Status :: 5 - Production/Stable",
        "Intended Audience :: Developers",
        "License :: OSI Approved :: GNU General Public License v3 or later (GPLv3+)",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: 3.10",
        "Programming Language :: Python :: 3.11",
        "Programming Language :: Python :: 3.12",
        "Programming Language :: Python :: 3.13",
    ],
    "Project-URL": [
        "Homepage, https://github.com/P403n1x87/austin",
        "Source Code, https://github.com/P403n1x87/austin",
        "Bug Tracker, https://github.com/P403n1x87/austin/issues",
        "Changelog, https://github.com/P403n1x87/austin/blob/master/ChangeLog",
        "Documentation, https://github.com/P403n1x87/austin/blob/master/README.md",
        "Funding, https://github.com/sponsors/P403n1x87",
        "Release Notes, https://github.com/P403n1x87/austin/releases",
    ],
    "Description-Content-Type": "text/markdown",
}


AUSTIN_WHEELS = {
    "manylinux_2_17_aarch64.manylinux2014_aarch64": (
        "gnu-linux-aarch64.tar.xz",
        ("austin", "austinp"),
    ),
    "manylinux_2_12_x86_64.manylinux2010_x86_64": (
        "gnu-linux-amd64.tar.xz",
        ("austin", "austinp"),
    ),
    "manylinux_2_17_armv7l.manylinux2014_armv7l": (
        "gnu-linux-armv7.tar.xz",
        ("austin", "austinp"),
    ),
    "manylinux_2_17_ppc64le.manylinux2014_ppc64le": (
        "gnu-linux-ppc64le.tar.xz",
        ("austin", "austinp"),
    ),
    "macosx_11_0_x86_64": ("mac64.zip", ("austin",)),
    "macosx_11_0_arm64": ("mac-arm64.zip", ("austin",)),
    "musllinux_1_1_aarch64": ("musl-linux-aarch64.tar.xz", ("austin",)),
    "musllinux_1_1_x86_64": ("musl-linux-amd64.tar.xz", ("austin",)),
    "musllinux_1_1_armv7l": ("musl-linux-armv7.tar.xz", ("austin",)),
    "musllinux_1_1_ppc64le": ("musl-linux-ppc64le.tar.xz", ("austin",)),
    "win_amd64": ("win64.zip", ("austin",)),
}


def make_message(headers, payload=None):
    message = StringIO()

    for name, value in headers.items():
        if isinstance(value, list):
            for value_part in value:
                print(f"{name}: {value_part}", file=message)
        else:
            print(f"{name}: {value}", file=message)

    if payload:
        print(file=message)
        print(payload, file=message)

    return message.getvalue().encode("utf-8")


def write_austin_wheel(out_dir, *, version, platform, austin_bin_data):
    package_name = "austin-dist"
    python = ".".join(("py2", "py3"))
    dist_name = package_name.replace("-", "_")
    wheel_name = f"{dist_name}-{version}-{python}-none-{platform}.whl"
    dist_info = f"{dist_name}-{version}.dist-info"
    wheel_path = out_dir / wheel_name

    print(f"Building {wheel_name} ... ", end="", flush=True)

    contents = {}

    for binary_name, binary_data in austin_bin_data:
        zip_info = ZipInfo(f"{dist_name}-{version}.data/scripts/{binary_name}")
        zip_info.external_attr |= 33261 << 16
        contents[zip_info] = binary_data

    contents[f"{dist_info}/METADATA"] = make_message(
        {
            "Metadata-Version": "2.1",
            "Name": package_name,
            "Version": version,
            **METADATA,
        },
        Path("README.md").read_text("utf-8"),
    )
    contents[f"{dist_info}/WHEEL"] = make_message(
        {
            "Wheel-Version": "1.0",
            "Generator": "austin-dist build-wheel.py",
            "Root-Is-Purelib": "false",
            "Tag": f"{python}-none-{platform}",
        }
    )

    with WheelFile(str(wheel_path), "w") as wheel:
        for member_info, member_source in contents.items():
            if not isinstance(member_info, ZipInfo):
                member_info = ZipInfo(member_info)
                member_info.external_attr = 0o644 << 16
            member_info.file_size = len(member_source)
            member_info.compress_type = ZIP_DEFLATED
            wheel.writestr(member_info, bytes(member_source))

    assert wheel_path.exists(), "wheel file created"

    print("done")


def get_latest_release() -> str:
    with urlopen(
        "https://api.github.com/repos/p403n1x87/austin/releases/latest"
    ) as stream:
        return json.loads(stream.read().decode("utf-8"))["tag_name"].strip("v")


def download_release(
    version: str, suffix: str, variant_name: str = "austin"
) -> tuple[str, bytes]:
    prefix = "https://github.com/p403n1x87/austin/releases/download/"
    try:
        with urlopen(f"{prefix}v{version}/{variant_name}-{version}-{suffix}") as stream:
            buffer = BytesIO(stream.read())
            if suffix.endswith(".tar.xz"):
                with tarfile.open(fileobj=buffer, mode="r:xz") as tar:
                    return variant_name, tar.extractfile(variant_name).read()
            elif suffix.endswith(".zip"):
                with ZipFile(buffer) as zip:
                    try:
                        return variant_name, zip.read(variant_name)
                    except KeyError:
                        file_name = f"{variant_name}.exe"
                        return file_name, zip.read(file_name)
            raise ValueError(f"Unknown archive extension: {suffix}")
    except HTTPError:
        raise RuntimeError(f"Could not download Austin version {version}")


if __name__ == "__main__":
    argp = ArgumentParser()

    argp.add_argument(
        "--version",
        help="Austin version to build wheels for",
        default=None,
    )

    argp.add_argument(
        "--platform",
        help="Platform to build wheels for",
        choices=AUSTIN_WHEELS.keys(),
        default=None,
    )

    argp.add_argument(
        "--files",
        help="The variant to binary file mapping (e.g. austin:src/austin)",
        nargs="+",
        type=str,
        default=None,
    )

    args = argp.parse_args()

    version = args.version or get_latest_release()

    dist_dir = Path.cwd() / "dist"
    dist_dir.mkdir(exist_ok=True)

    for platform, (suffix, variants) in AUSTIN_WHEELS.items():
        if args.platform is not None and args.platform != platform:
            continue

        bin_data = (
            (
                download_release(version, suffix, variant_name=variant)
                for variant in variants
            )
            if args.files is None
            else (
                (file_name, Path(bin_path).read_bytes())
                for file_name, bin_path in (file.split(":") for file in args.files)
            )
        )

        write_austin_wheel(
            dist_dir,
            version=version,
            platform=platform,
            austin_bin_data=bin_data,
        )



# This file is part of "austin" which is released under GPL.
#
# See file LICENCE or go to http://www.gnu.org/licenses/ for full license
# details.
#
# Austin is a Python frame stack sampler for CPython.
#
# Copyright (c) 2022 Gabriele N. Tornetta <phoenix1987@gmail.com>.
# All rights reserved.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import os
import platform
from collections import Counter
from shutil import rmtree
import time
from test.utils import allpythons
from test.utils import austin
from test.utils import austinp
from test.utils import compress
from test.utils import has_pattern
from test.utils import metadata
from test.utils import mojo
from test.utils import requires_sudo
from test.utils import run_python
from test.utils import sum_metric
from test.utils import target
from test.utils import threads
from test.utils import variants
from time import sleep

from flaky import flaky
import pytest


@requires_sudo
@pytest.mark.parametrize("heap", [tuple(), ("-h", "0"), ("-h", "64")])
@pytest.mark.parametrize(
    "mode,mode_meta", [("-i", "wall"), ("-si", "cpu"), ("-Ci", "wall"), ("-Csi", "cpu")]
)
@allpythons()
@variants
def test_attach_wall_time(austin, py, mode, mode_meta, heap):
    with run_python(py, target("sleepy.py"), "2") as p:
        sleep(0.5)

        result = austin(mode, "2ms", *heap, "-p", str(p.pid))
        assert result.returncode == 0

        ts = threads(result.stdout)
        assert len(ts) == 1, compress(result.stdout)

        assert has_pattern(result.stdout, "sleepy.py:<module>:"), compress(
            result.stdout
        )

        meta = metadata(result.stdout)

        assert meta["mode"] == mode_meta

        a = sum_metric(result.stdout)
        d = int(meta["duration"])

        assert a <= d


@requires_sudo
@pytest.mark.parametrize("exposure", [1, 2])
@allpythons()
def test_attach_exposure(py, exposure):
    with run_python(py, target("sleepy.py"), "3") as p:
        result = austin("-i", "1ms", "-x", str(exposure), "-p", str(p.pid))
        assert result.returncode == 0

        assert has_pattern(result.stdout, "sleepy.py:<module>:"), compress(
            result.stdout
        )

        meta = metadata(result.stdout)

        d = int(meta["duration"])

        assert exposure * 800000 <= d < exposure * 1200000

        p.kill()


@requires_sudo
@allpythons()
@mojo
def test_where(py, mojo):
    with run_python(py, target("sleepy.py"), sleep_after=1) as p:
        result = austin("-w", str(p.pid), mojo=mojo)
        assert result.returncode == 0

        assert "Process" in result.stdout
        assert "Thread" in result.stdout
        assert "sleepy.py" in result.stdout
        assert "<module>" in result.stdout


@flaky
@requires_sudo
@pytest.mark.xfail(platform.system() == "Windows", reason="Does not pass in Windows CI")
@allpythons()
def test_where_multiprocess(py):
    with run_python(py, target("target_mp.py")) as p:
        end = time.time() + 60
        while p.returncode is None and time.time() < end:
            sleep(0.2)
            result = austin("-Cw", str(p.pid))
            assert result.returncode == 0

            lines = Counter(result.stdout.splitlines())

            if sum(c for line, c in lines.items() if "Process" in line) >= 3:
                break
        else:
            raise AssertionError(compress(result.stdout))

        assert sum(c for line, c in lines.items() if "fact" in line) == 2, compress(
            result.stdout
        )
        (join_line,) = (line for line in lines if "join" in line)
        assert lines[join_line] == 1, compress(result.stdout)


@pytest.mark.xfail(reason="Fails in CI with some Python versions")
@requires_sudo
@allpythons()
def test_where_kernel(py):
    with run_python(py, target("sleepy.py"), sleep_after=1) as p:
        result = austinp("-kw", str(p.pid))
        assert result.returncode == 0

        assert "Process" in result.stdout, compress(result.stdout)
        assert "Thread" in result.stdout, compress(result.stdout)
        assert "sleepy.py" in result.stdout, compress(result.stdout)
        assert "<module>" in result.stdout, compress(result.stdout)
        assert "libc" in result.stdout, compress(result.stdout)
        assert "do_syscall" in result.stdout, compress(result.stdout)


@pytest.mark.parametrize("prefix", [[], ["unshare", "-p", "-f", "-r"]])
@pytest.mark.skipif(platform.system() != "Linux", reason="Linux only")
@requires_sudo
@allpythons()
def test_attach_container_like(py, tmp_path, prefix):
    """Test in container-like conditions.

    We test that we can still attach Austin to a running process even if we
    don't have access to the binary files to determine the location of the
    symbols. We also test against an interpreter started in a different PID
    namespace to emulate a container as closely as possible.
    """
    venv_path = tmp_path / ".venv"
    p = run_python(py, "-m", "venv", "--copies", str(venv_path))
    p.wait(30)
    assert 0 == p.returncode, "Virtual environment was created successfully"

    env = os.environ.copy()
    env["LD_LIBRARY_PATH"] = str(venv_path / "lib")
    env["PATH"] = str(venv_path / "bin") + os.pathsep + env["PATH"]
    with run_python(
        py, target("sleepy.py"), "3", env=env, prefix=prefix, sleep_after=0.5
    ) as p:
        rmtree(venv_path)
        sleep(0.5)

        result = austin("-Cp", str(p.pid))
        assert result.returncode == 0

        ts = threads(result.stdout)
        assert len(ts) == 1, compress(result.stdout)

        assert has_pattern(result.stdout, "sleepy.py:<module>:"), compress(
            result.stdout
        )

        meta = metadata(result.stdout)

        a = sum_metric(result.stdout)
        d = int(meta["duration"])

        assert abs(a - d) <= (a + d) * 0.25



import test.cunit.stats as stats
import time


def test_stats_gettime():
    begin = stats.gettime()
    time.sleep(1)
    delta = stats.gettime() - begin
    assert 0.95 <= delta / 1e6 <= 1.05


def test_stats_duration():
    stats.stats_start()
    time.sleep(1)
    dur = stats.stats_duration()
    assert 0.95 <= dur / 1e6 <= 1.05



import os
import sys
from contextlib import contextmanager
from pathlib import Path
from subprocess import PIPE
from subprocess import Popen
from subprocess import check_output
from tempfile import TemporaryDirectory
from test.utils import allpythons
from test.utils import austin
from test.utils import compress
from test.utils import has_pattern
from test.utils import requires_sudo
from test.utils import run_python
from test.utils import threads
from threading import Thread
from time import sleep

import psutil
import pytest
from requests import get


pytestmark = pytest.mark.skipif(
    sys.platform == "win32", reason="Not supported on Windows"
)

UWSGI = Path(__file__).parent / "uwsgi"


@contextmanager
def uwsgi(app="app.py", port=9090, args=[], env=None):
    with Popen(
        [
            "uwsgi",
            "--http",
            f":{port}",
            "--wsgi-file",
            UWSGI / app,
            *args,
        ],
        stdout=PIPE,
        stderr=PIPE,
        env=env or os.environ,
    ) as uw:
        sleep(0.5)

        assert uw.poll() is None, uw.stderr.read().decode()

        pid = uw.pid
        assert pid is not None, uw.stderr.read().decode()

        try:
            yield uw
        finally:
            uw.kill()
            # Sledgehammer
            for proc in psutil.process_iter():
                if "uwsgi" in proc.name():
                    proc.terminate()
                    proc.wait()


@contextmanager
def venv(py, reqs):
    with TemporaryDirectory() as tmp_path:
        venv_path = Path(tmp_path) / ".venv"
        p = run_python(py, "-m", "venv", str(venv_path))
        p.wait(120)
        assert 0 == p.returncode, "Virtual environment was created successfully"

        env = os.environ.copy()
        env["LD_LIBRARY_PATH"] = str(venv_path / "lib")
        env["PATH"] = str(venv_path / "bin") + os.pathsep + env["PATH"]

        check_output(
            ["python", "-m", "pip", "install", "-r", reqs, "--use-pep517"], env=env
        )

        yield env


@requires_sudo
@allpythons()
def test_uwsgi(py):
    request_thread = Thread(target=get, args=("http://localhost:9090",))

    with venv(py, reqs=Path(__file__).parent / "requirements-uwsgi.txt") as env:
        with uwsgi(env=env) as uw:
            request_thread.start()

            result = austin("-x", "2", "-Cp", str(uw.pid))
            assert has_pattern(result.stdout, "app.py:application:5"), compress(
                result.stdout
            )

        request_thread.join()


@requires_sudo
@allpythons()
def test_uwsgi_multiprocess(py):
    request_thread = Thread(target=get, args=("http://localhost:9091",))

    with venv(py, reqs=Path(__file__).parent / "requirements-uwsgi.txt") as env:
        with uwsgi(
            port=9091, args=["--processes", "2", "--threads", "2"], env=env
        ) as uw:
            request_thread.start()

            result = austin("-x", "2", "-Cp", str(uw.pid))
            assert has_pattern(result.stdout, "app.py:application:5"), compress(
                result.stdout
            )

            ts = threads(result.stdout)
            assert len(ts) >= 4, ts
            assert len({p for p, _ in ts}) >= 2, ts

        request_thread.join()



# Run as python3 scripts/benchmark.py from the repository root directory.
# Ensure dependencies from requirements-bm.txt are installed.

import abc
import re
import sys
import typing as t
from argparse import ArgumentParser
from math import floor, log
from pathlib import Path
from test.utils import metadata, target
from textwrap import wrap

from common import download_release
from scipy.stats import ttest_ind

VERSIONS = ("3.5.0", "3.6.0", "dev")
SCENARIOS = [
    *[
        (
            "austin",
            f"Wall time [sampling interval: {i}]",
            ["-Pi", str(i), sys.executable, target("target34.py")],
        )
        for i in (1, 10, 100, 1000)
    ],
    *[
        (
            "austin",
            f"CPU time [sampling interval: {i}]",
            ["-Psi", str(i), sys.executable, target("target34.py")],
        )
        for i in (1, 10, 100, 1000)
    ],
    *[
        (
            "austin",
            f"RSA keygen [sampling interval: {i}]",
            ["-Psi", str(i), sys.executable, "-m", "test.bm.rsa_key_generator"],
        )
        for i in (1, 10, 100, 1000)
    ],
    *[
        (
            "austin",
            f"Full metrics [sampling interval: {i}]",
            ["-Pfi", str(i), sys.executable, target("target34.py")],
        )
        for i in (1, 10, 100, 1000)
    ],
    *[
        (
            "austin",
            f"Multiprocess wall time [sampling interval: {i}]",
            ["-CPfi", str(i), sys.executable, target("target_mp.py"), "16"],
        )
        for i in (1, 10, 100, 1000)
    ],
]


# The metrics we evaluate and whether they are to be maximised or minimised.
METRICS = [
    ("Sample Rate", +1),
    ("Saturation", -1),
    ("Error Rate", -1),
    ("Sampling Speed", -1),
]


def get_stats(output: str) -> t.Optional[dict]:
    try:
        meta = metadata(output)
        raw_saturation = meta["saturation"]
        _, _, raw_samples = raw_saturation.partition("/")

        duration = float(meta["duration"]) / 1e6
        samples = int(raw_samples)
        saturation = eval(raw_saturation)
        error_rate = eval(meta["errors"])
        sampling = int(meta["sampling"].split(",")[1])

        return {
            "Sample Rate": samples / duration,
            "Saturation": saturation,
            "Error Rate": error_rate,
            "Sampling Speed": sampling,
        }

    except Exception:
        # Failed to get stats
        return None


class Outcome:
    __critical_p__ = 0.025

    def __init__(self, data: list[float]) -> None:
        self.data = data
        self.mean = sum(data) / len(data)
        self.stdev = (
            sum(((v - self.mean) ** 2 for v in data)) / (len(data) - 1)
        ) ** 0.5

    def __repr__(self):
        n = -floor(log(self.stdev, 10)) if self.stdev else 0
        rmean = round(self.mean, n)
        rstdev = round(self.stdev, n)
        if n <= 0:
            rmean = int(rmean)
            rstdev = int(rstdev)

        return f"{rmean} ± {rstdev}"

    __str__ = __repr__

    def __len__(self):
        return len(repr(self))

    def __eq__(self, other: "Outcome") -> bool:
        t, p = ttest_ind(self.data, other.data, equal_var=False)
        return p < self.__critical_p__


Results = t.Tuple[str, t.Dict[str, Outcome]]


class Renderer(abc.ABC):
    BETTER = "better"
    WORSE = "worse"
    SAME = "same"

    @abc.abstractmethod
    def render_header(self, title: str, level: int = 1) -> str:
        ...

    @abc.abstractmethod
    def render_paragraph(self, text: str) -> str:
        ...

    @abc.abstractmethod
    def render_table(self, table) -> str:
        ...

    @abc.abstractmethod
    def render_scenario(
        self, title, results: t.List[t.Tuple[str, t.List[Results]]]
    ) -> str:
        ...

    @abc.abstractmethod
    def render_summary(
        self, summary: t.List[t.Tuple[str, t.List[t.Tuple[str, bool, int]]]]
    ) -> str:
        ...

    def render_scenario(
        self, title, table: t.List[t.Tuple[str, t.List[Results]]]
    ) -> str:
        self.render_header(title, level=2)
        self.render_table(table)
        print()

    def render_summary(self, summary):
        self.render_header("Benchmark Summary", level=2)
        self.render_paragraph(f"Comparison of {VERSIONS[-1]} against {VERSIONS[-2]}.")

        if not summary:
            self.render_paragraph(
                "No significant difference in performance between versions."
            )
            return

        self.render_paragraph(
            "The following scenarios show a statistically significant difference "
            "in performance between the two versions."
        )

        self.render_table(
            [
                (
                    title,
                    {
                        m: {1: self.BETTER, -1: self.WORSE}[s] if c else self.SAME
                        for m, c, s in tests
                    },
                )
                for title, tests in summary
            ]
        )


class TerminalRenderer(Renderer):
    def render_table(self, table: t.List[t.Tuple[str, t.List[Results]]]) -> str:
        _, row = table[0]
        cols = list(row.keys())
        max_vh = max(len(e[0]) for e in table)

        col_widths = [max(max(len(r[col]), len(col)) for _, r in table) for col in cols]
        div_len = sum(col_widths) + (len(cols) + 1) * 2 + max_vh

        print("=" * div_len)
        print(
            (" " * (max_vh + 2))
            + "".join(f"{col:^{cw+2}}" for col, cw in zip(cols, col_widths))
        )
        print("-" * div_len)

        for v, row in table:
            print(f"{v:^{max_vh+2}}", end="")
            for col, cw in zip(cols, col_widths):
                print(f"{str(row[col]):^{cw+2}}", end="")
            print()

        print("=" * div_len)

    def render_header(self, title: str, level: int = 1) -> str:
        print(title)
        print({1: "=", 2: "-", 3: "~"}.get(level, "-") * len(title))
        print()

    def render_paragraph(self, text: str) -> str:
        for _ in wrap(text):
            print(_)
        print()


class MarkdownRenderer(Renderer):
    BETTER = ":green_circle:"
    WORSE = ":red_circle:"
    SAME = ":yellow_circle:"

    def render_header(self, title: str, level: int = 1) -> str:
        print(f"{'#' * level} {title}")
        print()

    def render_paragraph(self, text: str) -> str:
        print(text)
        print()

    def render_table(self, table: t.List[t.Tuple[str, t.List[Results]]]) -> str:
        _, row = table[0]
        cols = list(row.keys())
        max_vh = max(len(e[0]) for e in table)

        col_widths = [max(max(len(r[col]), len(col)) for _, r in table) for col in cols]
        div_len = sum(col_widths) + (len(cols) + 1) * 2 + max_vh

        print("|     |" + "|".join(f" {col} " for col in cols) + "|")
        print("| --- |" + "|".join(f":{'-' * len(col)}:" for col in cols) + "|")

        for v, row in table:
            print(
                f"| {v} |"
                + "|".join(
                    f" {str(row[col]):^{cw}} " for col, cw in zip(cols, col_widths)
                )
                + "|"
            )

    def render_scenario(
        self, title, table: t.List[t.Tuple[str, t.List[Results]]]
    ) -> str:
        print("<details>")
        print(f"<summary><strong>{title}</strong></summary>")
        print()
        super().render_scenario(title, table)
        print("</details>")
        print()


def summarize(results: t.List[t.Tuple[str, t.List[Results]]]):
    summary = []
    for title, table in results:
        (_, a), (_, b) = table[-2:]
        tests = [
            (
                m,
                a[m] == b[m],
                int((b[m].mean - a[m].mean) * s / (abs(b[m].mean - a[m].mean) or 1)),
            )
            for m, s in METRICS
        ]
        if any(c for _, c, _ in tests):
            summary.append((title, tests))
    return summary


def main():
    argp = ArgumentParser()

    argp.add_argument(
        "-k",
        type=re.compile,
        help="Run benchmark scenarios that match the given regular expression",
    )

    argp.add_argument(
        "-n",
        type=int,
        default=10,
        help="Number of times to run each scenario",
    )

    argp.add_argument(
        "-f",
        "--format",
        type=str,
        choices=["terminal", "markdown"],
        default="terminal",
        help="The output format",
    )

    argp.add_argument(
        "-l",
        "--last",
        action="store_true",
        help="Run only with the last release of Austin",
    )

    argp.add_argument(
        "-p",
        "--pvalue",
        type=float,
        default=0.025,
        help="The p-value to use when testing for statistical significance",
    )

    opts = argp.parse_args()

    Outcome.__critical_p__ = opts.pvalue

    renderer = {"terminal": TerminalRenderer, "markdown": MarkdownRenderer}[
        opts.format
    ]()

    renderer.render_header("Austin Benchmarks")
    renderer.render_paragraph(
        f"Running Austin benchmarks with Python {'.'.join(str(_) for _ in sys.version_info[:3])}",
    )

    results: t.List[t.Tuple[str, t.List[Results]]] = []

    for variant, title, args in SCENARIOS:
        if opts.k is not None and not opts.k.match(title):
            continue

        table: t.List[Results] = []
        for version in VERSIONS[-2:] if opts.last else VERSIONS:
            print(f"> Running with Austin {version} ...    ", end="\r", file=sys.stderr)
            try:
                austin = download_release(version, Path("/tmp"), variant_name=variant)
            except RuntimeError:
                print(
                    f"WARNING: Could not download {variant} {version}", file=sys.stderr
                )
                continue

            stats = [
                _
                for _ in (get_stats(austin(*args).stdout) for _ in range(opts.n))
                if _ is not None
            ]
            table.append(
                (
                    version,
                    {
                        key: Outcome([s[key] for s in stats])
                        for key in list(stats[0].keys())
                    },
                )
            )

        results.append((title, table))

    summary = summarize(results)

    renderer.render_summary(summary)

    renderer.render_header("Benchmark Results", level=2)
    for title, table in results:
        renderer.render_scenario(title, table)


if __name__ == "__main__":
    try:
        main()
    except KeyboardInterrupt:
        print("\nBye!")



from test.cunit.error import cglobal
from test.cunit.error import error_get_msg
from test.cunit.error import is_fatal

import pytest


@pytest.mark.parametrize(
    "code, msg",
    [
        (0, "No error"),
        (1, "Cannot open memory maps file"),
        (8, "Failed to retrieve PyCodeObject"),
        (16, "Failed to create frame object"),
        (24, "Failed to create thread object"),
        (32, "Failed to initialise process"),
    ],
)
def test_error_get_msg(code, msg):
    assert error_get_msg(code) == msg.encode()


def test_error_unknown():
    assert error_get_msg(10000) == b"<Unknown error>"


@pytest.mark.parametrize(
    "code, fatal",
    [
        (0, False),
        (1, True),
        (8, False),
        (32, False),
        (1000, False),
    ],
)
def test_error_is_fatal(code, fatal):
    assert is_fatal(code) == fatal


def test_error_austin_errno_global():
    assert cglobal("austin_errno", "int") == 0



# This file is part of "austin" which is released under GPL.
#
# See file LICENCE or go to http://www.gnu.org/licenses/ for full license
# details.
#
# Austin is a Python frame stack sampler for CPython.
#
# Copyright (c) 2022 Gabriele N. Tornetta <phoenix1987@gmail.com>.
# All rights reserved.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import importlib
import os
import platform
from asyncio.subprocess import STDOUT
from collections import Counter
from collections import defaultdict
from io import BytesIO
from io import StringIO
from pathlib import Path
from shutil import rmtree
from subprocess import PIPE
from subprocess import CalledProcessError
from subprocess import CompletedProcess
from subprocess import Popen
from subprocess import TimeoutExpired
from subprocess import check_output
from tempfile import gettempdir
from test import PYTHON_VERSIONS
from time import sleep
from types import ModuleType
from typing import Iterator
from typing import List
from typing import TypeVar
from typing import Union


try:
    import pytest
except ImportError:
    pytest = None

from austin.format.mojo import MojoFile


HERE = Path(__file__).parent


def target(name: str = "target34.py") -> str:
    return str(HERE / "targets" / name)


def allpythons(min=None, max=None):
    def _(f):
        versions = PYTHON_VERSIONS
        if min is not None:
            versions = [_ for _ in versions if _ >= min]
        if max is not None:
            versions = [_ for _ in versions if _ <= max]
        return pytest.mark.parametrize(
            "py", [".".join([str(_) for _ in v]) for v in versions]
        )(f)

    return _


if platform.system() == "Darwin":
    BREW_PREFIX = check_output(["brew", "--prefix"], text=True, stderr=STDOUT).strip()


def python(version: str) -> list[str]:
    match platform.system():
        case "Windows":
            py = ["py", f"-{version}"]
        case "Darwin" | "Linux":
            py = [f"python{version}"]
        case _:
            raise RuntimeError(f"Unsupported platform: {platform.system()}")

    try:
        check_output([*py, "-V"], stderr=STDOUT)
        return py
    except FileNotFoundError:
        pytest.skip(f"Python {version} not found")


def gdb(cmds: list[str], *args: tuple[str]) -> str:
    return check_output(
        ["gdb", "-q", "-batch"]
        + [_ for cs in (("-ex", _) for _ in cmds) for _ in cs]
        + list(args),
        stderr=STDOUT,
    ).decode()


def apport_unpack(report: Path, target_dir: Path):
    return check_output(
        ["apport-unpack", str(report), str(target_dir)],
        stderr=STDOUT,
    ).decode()


def bt(binary: Path, pid: int) -> str:
    core_file = f"core.{pid}"
    if Path(core_file).is_file():
        return gdb(["bt full", "q"], str(binary), core_file)

    # On Ubuntu, apport puts crashes in /var/crash
    crash_dir = Path("/var/crash")
    if crash_dir.is_dir():
        crashes = list(crash_dir.glob("*.crash"))
        if crashes:
            # Take the last one
            crash = crashes[-1]
            target_dir = Path(crash.stem)
            apport_unpack(crash, target_dir)

            result = gdb(["bt full", "q"], str(binary), target_dir / "CoreDump")

            crash.unlink()
            rmtree(str(target_dir))

            return result

    return "No core dump available."


def collect_logs(variant: str, pid: int) -> List[str]:
    match platform.system():
        case "Linux":
            with Path("/var/log/syslog").open() as logfile:
                needles = (f"{variant}[{pid}]", f"systemd-coredump[{pid}]")
                return [
                    f" logs for {variant}[{pid}] ".center(80, "="),
                    *(
                        line.strip().replace("#012", "\n")
                        for line in logfile.readlines()
                        if any(needle in line for needle in needles)
                    ),
                    f" end of logs for {variant}[{pid}] ".center(80, "="),
                ]

        case "Windows":
            with (Path(gettempdir()) / "austin.log").open() as logfile:
                needles = (f"{variant}[{pid}]",)
                return [
                    f" logs for {variant}[{pid}] ".center(80, "="),
                    *(
                        line.strip()
                        for line in logfile.readlines()
                        if any(needle in line for needle in needles)
                    ),
                    f" end of logs for {variant}[{pid}] ".center(80, "="),
                ]

        case _:
            return []


EXEEXT = ".exe" if platform.system() == "Windows" else ""


# Taken from the subprocess module. We make our own version that can also
# propagate the PID.
def run(
    *popenargs, input=None, capture_output=False, timeout=None, check=False, **kwargs
):
    if input is not None:
        if kwargs.get("stdin") is not None:
            raise ValueError("stdin and input arguments may not both be used.")
        kwargs["stdin"] = PIPE

    if capture_output:
        if kwargs.get("stdout") is not None or kwargs.get("stderr") is not None:
            raise ValueError(
                "stdout and stderr arguments may not be used " "with capture_output."
            )
        kwargs["stdout"] = PIPE
        kwargs["stderr"] = PIPE

    with Popen(*popenargs, **kwargs) as process:
        try:
            stdout, stderr = process.communicate(input, timeout=timeout)
        except TimeoutExpired as exc:
            exc.pid = process.pid
            process.kill()
            if platform.system() == "Windows":
                exc.stdout, exc.stderr = process.communicate()
            else:
                process.wait()
            raise
        except Exception as e:
            e.pid = process.pid
            process.kill()
            raise
        retcode = process.poll()
        if check and retcode:
            exc = CalledProcessError(
                retcode, process.args, output=stdout, stderr=stderr
            )
            exc.pid = process.pid
            raise exc
    result = CompletedProcess(process.args, retcode, stdout, stderr)
    result.pid = process.pid

    return result


def print_logs(logs: List[str]) -> None:
    if logs:
        for log in logs:
            print(log)
    else:
        print("<< no logs available >>")


class Variant(str):
    ALL: list["Variant"] = []

    def __init__(self, name: str) -> None:
        super().__init__()

        path = (Path("src") / name).with_suffix(EXEEXT)
        if not path.is_file():
            path = Path(name).with_suffix(EXEEXT)

        self.name = name
        self.path = path

        self.ALL.append(self)

    def __call__(
        self,
        *args: str,
        timeout: int = 60,
        mojo: bool = False,
        convert: bool = True,
        expect_fail: Union[bool, int] = False,
    ) -> CompletedProcess:
        if not self.path.is_file():
            pytest.skip(f"Variant '{self}' not available")

        mojo_args = ["-b"] if mojo else []

        try:
            result = run(
                [str(self.path)] + mojo_args + list(args),
                capture_output=True,
                timeout=timeout,
            )
        except Exception as exc:
            if (pid := getattr(exc, "pid", None)) is not None:
                print_logs(collect_logs(self.name, pid))
            raise

        if result.returncode in (-11, 139):  # SIGSEGV
            print(bt(self.path, result.pid))

        if mojo and not ({"-o", "-w", "--output", "--where"} & set(args)):
            # We produce MOJO binary data only if we are not writing to file
            # or using the "where" option.
            if convert:
                result.stdout = demojo(result.stdout)
        else:
            result.stdout = result.stdout.decode(errors="ignore")
        result.stderr = result.stderr.decode()

        logs = collect_logs(self.name, result.pid)
        result.logs = logs

        if result.returncode != int(expect_fail):
            print_logs(logs)

        return result


austin = Variant("austin")
austinp = Variant("austinp")


def run_async(command: list[str], *args: tuple[str], env: dict | None = None) -> Popen:
    return Popen(command + list(args), stdout=PIPE, stderr=PIPE, env=env)


def run_python(
    version,
    *args: tuple[str],
    env: dict | None = None,
    prefix: list[str] = [],
    sleep_after: float | None = None,
) -> Popen:
    result = run_async(prefix + python(version), *args, env=env)

    if sleep_after is not None:
        sleep(sleep_after)

    return result


def samples(data: str) -> Iterator[bytes]:
    return (_ for _ in data.splitlines() if _ and _[0] == "P")


T = TypeVar("T")


def denoise(data: Iterator[T], threshold: float = 0.1) -> set[T]:
    c = Counter(data)
    try:
        m = max(c.values())
    except ValueError:
        return set()
    return {t for t, c in c.items() if c / m > threshold}


def processes(data: str) -> set[str]:
    return denoise(_.partition(";")[0] for _ in samples(data))


def threads(data: str, threshold: float = 0.1) -> set[tuple[str, str]]:
    return denoise(
        tuple(_.rpartition(" ")[0].split(";", maxsplit=2)[:2]) for _ in samples(data)
    )


def metadata(data: str) -> dict[str, str]:
    meta = dict(
        _[1:].strip().split(": ", maxsplit=1)
        for _ in data.splitlines()
        if _ and _[0] == "#" and not _.startswith("# map:")
    )

    for v in ("austin", "python"):
        if v in meta:
            meta[v] = tuple(int(_.replace("?", "-1")) for _ in meta[v].split(".")[:3])

    return meta


def maps(data: str) -> defaultdict[str, list[str]]:
    maps = defaultdict(list)
    for r, f in (
        _[7:].split(" ", maxsplit=1)
        for _ in data.splitlines()
        if _.startswith("# map:")
    ):
        maps[f].append(r)
    return maps


def has_pattern(data: str, pattern: str) -> bool:
    for _ in samples(data):
        if pattern in _:
            return True
    return False


def sum_metric(data: str) -> int:
    return sum(int(_.rpartition(" ")[-1]) for _ in samples(data))


def sum_metrics(data: str) -> tuple[int, int, int, int]:
    wall = cpu = alloc = dealloc = 0
    for t, i, m in (
        _.rpartition(" ")[-1].split(",", maxsplit=2) for _ in samples(data)
    ):
        time = int(t)
        wall += time
        if i == "0":
            cpu += time

        memory = int(m)
        if memory > 0:
            alloc += memory
        else:
            dealloc += memory

    return wall, cpu, alloc, dealloc


def compress(data: str) -> str:
    stacks: dict[str, int] = {}

    for _ in (_.strip() for _ in data.splitlines() if _ and _[0] == "P"):
        stack, _, metric = _.rpartition(" ")
        stacks[stack] = stacks.setdefault(stack, 0) + int(metric)

    compressed_stacks = "\n".join((f"{k} {v}" for k, v in stacks.items()))

    output = (
        f"# Metadata\n{metadata(data)}\n\n# Stacks\n{compressed_stacks or '<no data>'}"
    )

    ms = maps(data)
    if ms:
        output = f"# Maps\n{list(ms.keys())}\n\n" + output

    return output


def demojo(data: bytes) -> str:
    result = StringIO()

    for e in MojoFile(BytesIO(data)).parse():
        result.write(e.to_austin())

    return result.getvalue()


# Load from the utils scripts
def load_util(name: str) -> ModuleType:
    module_path = (Path(__file__).parent.parent / "utils" / name).with_suffix(".py")
    spec = importlib.util.spec_from_file_location(name, str(module_path))

    assert spec is not None and spec.loader is not None, spec

    module = importlib.util.module_from_spec(spec)
    spec.loader.exec_module(module)

    return module


if pytest is not None:
    variants = pytest.mark.parametrize("austin", Variant.ALL)

    match platform.system():
        case "Windows":
            requires_sudo = no_sudo = lambda f: f
        case _:
            requires_sudo = pytest.mark.skipif(
                os.geteuid() != 0, reason="Requires superuser privileges"
            )
            no_sudo = pytest.mark.skipif(
                os.geteuid() == 0, reason="Must not have superuser privileges"
            )

    mojo = pytest.mark.parametrize("mojo", [False, True])



import os
import sys
import typing as t
from pathlib import Path
from subprocess import PIPE
from test.cunit import SRC
from test.utils import bt
from test.utils import run
from types import FunctionType

import pytest


class SegmentationFault(Exception):
    pass


class CUnitTestFailure(Exception):
    pass


def pytest_configure(config: pytest.Config) -> None:
    # register an additional marker
    config.addinivalue_line("markers", "exitcode(code): the expected exit code")


def pytest_pycollect_makeitem(
    collector: pytest.Collector, name: str, obj: t.Any
) -> None:
    if (
        not os.getenv("PYTEST_CUNIT")
        and isinstance(obj, FunctionType)
        and name.startswith("test_")
    ):
        obj.__cunit__ = (str(collector.fspath), name)


def cunit(
    module: str, name: str, full_name: str, exit_code: int = 0
) -> t.Callable[[t.Any, t.Any], None]:
    def _(*_, **__):
        test = f"{module}::{name}"
        env = os.environ.copy()
        env["PYTEST_CUNIT"] = full_name

        result = run(
            [sys.executable, "-m", "pytest", "-svv", test],
            stdout=PIPE,
            stderr=PIPE,
            env=env,
        )

        if result.returncode == exit_code:
            return

        if result.returncode == -11:
            binary_name = Path(module).stem.replace("test_", "")
            raise SegmentationFault(
                bt((SRC / binary_name).with_suffix(".so"), result.pid)
            )

        raise CUnitTestFailure(
            f"\n{result.stdout.decode()}\n"
            f"Process terminated with exit code {result.returncode} "
            "(expected {exit_code})"
        )

    return _


def pytest_collection_modifyitems(
    session: pytest.Session,
    config: pytest.Config,
    items: list[pytest.Item],
) -> None:
    if test_name := os.getenv("PYTEST_CUNIT"):
        # We are inside the sandbox process. We select the only test we care
        items[:] = [_ for _ in items if _.name == test_name]
        return

    for item in items:
        if hasattr(item._obj, "__cunit__"):
            exit_code_marker = list(item.iter_markers(name="exitcode"))
            exit_code = exit_code_marker[0].args[0] if exit_code_marker else 0
            item._obj = cunit(
                *item._obj.__cunit__, full_name=item.name, exit_code=exit_code
            )



import test.cunit.argparse as argparse
from ctypes import c_char_p

import pytest


def parse_args(argv):
    argc = len(argv)
    return argparse.parse_args(argc, (c_char_p * argc)(*(_.encode() for _ in argv)))


def test_parse_args_command():
    assert parse_args(["austin", "python"])


# FIXME
@pytest.mark.exitcode(1)
def test_parse_args_process():
    assert parse_args(["austin", "-p", "123"])



# This file is part of "austin" which is released under GPL.
#
# See file LICENCE or go to http://www.gnu.org/licenses/ for full license
# details.
#
# Austin is a Python frame stack sampler for CPython.
#
# Copyright (c) 2022 Gabriele N. Tornetta <phoenix1987@gmail.com>.
# All rights reserved.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from subprocess import run
from test.utils import allpythons
from test.utils import austin
from test.utils import python
from test.utils import requires_sudo
from test.utils import run_python
from test.utils import target

import pytest


def valgrind(python_args: list[str], mode: str):
    try:
        return run(
            [
                "valgrind",
                "--error-exitcode=42",
                "--leak-check=full",
                "--show-leak-kinds=all",
                "--errors-for-leak-kinds=all",
                "--track-fds=yes",
                str(austin.path),
                f"-{mode}i",
                "1ms",
                "-o",
                "/dev/null",
                *python_args,
            ],
            capture_output=True,
            timeout=30,
            text=True,
        )
    except FileNotFoundError:
        pytest.skip("Valgrind not available")


@pytest.mark.parametrize("mode", ["", "s", "C", "Cs"])
@allpythons()
def test_valgrind_fork(py, mode):
    result = valgrind([*python(py), target()], mode)
    assert result.returncode == 0, "\n".join((result.stdout, result.stderr))


@requires_sudo
@pytest.mark.parametrize("mode", ["", "s", "C", "Cs"])
@allpythons()
def test_valgrind_attach(py, mode):
    with run_python(py, target("sleepy.py"), "2") as p:
        result = valgrind(["-p", str(p.pid)], mode)
        assert result.returncode == 0, "\n".join((result.stdout, result.stderr))
        p.kill()



# This file is part of "austin" which is released under GPL.
#
# See file LICENCE or go to http://www.gnu.org/licenses/ for full license
# details.
#
# Austin is a Python frame stack sampler for CPython.
#
# Copyright (c) 2022 Gabriele N. Tornetta <phoenix1987@gmail.com>.
# All rights reserved.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from test.utils import allpythons
from test.utils import austin
from test.utils import compress
from test.utils import has_pattern
from test.utils import python
from test.utils import samples
from test.utils import target

import pytest


@pytest.mark.parametrize("heap", [tuple(), ("-h", "0"), ("-h", "64")])
@allpythons()
def test_accuracy_fast_recursive(py, heap):
    result = austin("-i", "1ms", "-P", *heap, *python(py), target("recursive.py"))
    assert result.returncode == 0, result.stderr or result.stdout

    assert has_pattern(result.stdout, "sum_up_to"), compress(result.stdout)

    for _ in samples(result.stdout):
        if "sum_up_to" in _ and "<module>" in _:
            assert len(_.split(";")) <= 20 or has_pattern(_, ":INVALID:"), _



# This file is part of "austin" which is released under GPL.
#
# See file LICENCE or go to http://www.gnu.org/licenses/ for full license
# details.
#
# Austin is a Python frame stack sampler for CPython.
#
# Copyright (c) 2022 Gabriele N. Tornetta <phoenix1987@gmail.com>.
# All rights reserved.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from test.utils import allpythons
from test.utils import austin
from test.utils import has_pattern
from test.utils import metadata
from test.utils import mojo
from test.utils import python
from test.utils import samples
from test.utils import target


@allpythons()
@mojo
def test_gc_off(py, mojo):
    result = austin("-i", "1ms", *python(py), target("target_gc.py"), mojo=mojo)
    assert result.returncode == 0

    assert not has_pattern(":GC:", result.stdout)


@allpythons()
@mojo
def test_gc_on(py, mojo):
    result = austin("-gi", "1ms", *python(py), target("target_gc.py"), mojo=mojo)
    assert result.returncode == 0

    meta = metadata(result.stdout)
    assert float(meta["gc"]) / float(meta["duration"]) > 0.1

    gcs = [_ for _ in samples(result.stdout) if ":GC:" in _]
    assert len(gcs) > 10


@allpythons()
@mojo
def test_gc_disabled(py, monkeypatch, mojo):
    monkeypatch.setenv("GC_DISABLED", "1")

    result = austin("-gi", "10ms", *python(py), target("target_gc.py"), mojo=mojo)
    assert result.returncode == 0

    meta = metadata(result.stdout)
    assert int(meta["gc"]) * 0.8 < int(meta["duration"]) / 20

    gcs = [_ for _ in samples(result.stdout) if ":GC:" in _]
    assert len(gcs) < 5



# This file is part of "austin" which is released under GPL.
#
# See file LICENCE or go to http://www.gnu.org/licenses/ for full license
# details.
#
# Austin is a Python frame stack sampler for CPython.
#
# Copyright (c) 2022 Gabriele N. Tornetta <phoenix1987@gmail.com>.
# All rights reserved.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import platform
from pathlib import Path
from test.utils import allpythons
from test.utils import austin
from test.utils import compress
from test.utils import demojo
from test.utils import has_pattern
from test.utils import maps
from test.utils import metadata
from test.utils import mojo
from test.utils import processes
from test.utils import python
from test.utils import samples
from test.utils import sum_metric
from test.utils import sum_metrics
from test.utils import target
from test.utils import threads
from test.utils import variants

import pytest


@pytest.mark.parametrize("heap", [tuple(), ("-h", "0"), ("-h", "64")])
@allpythons()
@variants
@mojo
def test_fork_wall_time(austin, py, heap, mojo):
    result = austin("-i", "2ms", *heap, *python(py), target("target34.py"), mojo=mojo)
    assert py in (result.stderr or result.stdout), result.stderr or result.stdout

    assert len(processes(result.stdout)) == 1, compress(result.stdout)
    ts = threads(result.stdout)
    assert len(ts) == 2, compress(result.stdout)
    assert all(len(t[1].split(":")) == 2 for t in ts), "threads have interpreter ID"

    assert has_pattern(result.stdout, "target34.py:keep_cpu_busy:3"), compress(
        result.stdout
    )
    assert not has_pattern(result.stdout, "Unwanted")

    meta = metadata(result.stdout)

    assert meta["mode"] == "wall"

    a = sum_metric(result.stdout)
    d = int(meta["duration"])

    assert 0 < a < 2.1 * d

    if austin == "austinp":
        ms = maps(result.stdout)
        assert len(ms) >= 2, ms
        assert [_ for _ in ms if "python" in _], ms


@pytest.mark.parametrize("heap", [tuple(), ("-h", "0"), ("-h", "64")])
@allpythons()
@variants
@mojo
def test_fork_cpu_time_cpu_bound(py, heap, austin, mojo):
    result = austin("-si", "1ms", *heap, *python(py), target("target34.py"), mojo=mojo)
    assert result.returncode == 0, result.stderr or result.stdout

    assert has_pattern(result.stdout, "target34.py:keep_cpu_busy:3"), compress(
        result.stdout
    )
    assert not has_pattern(result.stdout, "Unwanted")

    meta = metadata(result.stdout)

    assert meta["mode"] == "cpu"

    a = sum_metric(result.stdout)
    d = int(meta["duration"])

    assert 0 < a < 2.1 * d


@allpythons()
@variants
def test_fork_cpu_time_idle(py, austin):
    result = austin("-si", "1ms", *python(py), target("sleepy.py"))
    assert result.returncode == 0, result.stderr or result.stdout

    assert has_pattern(result.stdout, "sleepy.py:<module>:"), compress(result.stdout)

    meta = metadata(result.stdout)

    a = sum_metric(result.stdout)
    d = int(meta["duration"])

    assert a < 1.1 * d


@allpythons()
@mojo
def test_fork_memory(py, mojo):
    result = austin("-mi", "1ms", *python(py), target("target34.py"), mojo=mojo)
    assert result.returncode == 0, result.stderr or result.stdout

    assert has_pattern(result.stdout, "target34.py:keep_cpu_busy:32")

    meta = metadata(result.stdout)

    assert meta["mode"] == "memory"

    d = int(meta["duration"])
    assert d > 100000

    ms = [int(_.rpartition(" ")[-1]) for _ in samples(result.stdout)]
    alloc = sum(_ for _ in ms if _ > 0)
    dealloc = sum(-_ for _ in ms if _ < 0)

    assert alloc * dealloc


@allpythons()
@mojo
def test_fork_output(py, tmp_path: Path, mojo):
    datafile = tmp_path / "test_fork_output.austin"

    result = austin(
        "-i", "1ms", "-o", str(datafile), *python(py), target("target34.py"), mojo=mojo
    )
    assert result.returncode == 0, result.stderr or result.stdout

    assert "Unwanted" in result.stdout

    data = demojo(datafile.read_bytes()) if mojo else datafile.read_text()

    assert has_pattern(data, "target34.py:keep_cpu_busy:32")

    meta = metadata(data)

    assert meta["mode"] == "wall"

    a = sum(int(_.rpartition(" ")[-1]) for _ in samples(data))
    d = int(meta["duration"])

    assert 0 < 0.9 * d < a < 2.1 * d


# Support for multiprocess is attach-like and seems to suffer from the same
# issues as attach tests on Windows.
@pytest.mark.xfail(platform.system() == "Windows", reason="Does not pass in Windows CI")
@allpythons()
@mojo
def test_fork_multiprocess(py, mojo):
    result = austin("-Ci", "1ms", *python(py), target("target_mp.py"), mojo=mojo)
    assert result.returncode == 0, result.stderr or result.stdout

    ps = processes(result.stdout)
    assert len(ps) >= 3, ps

    meta = metadata(result.stdout)
    assert meta["multiprocess"] == "on", meta
    assert meta["mode"] == "wall", meta

    assert has_pattern(result.stdout, "target_mp.py:do:"), compress(result.stdout)
    assert has_pattern(result.stdout, "target_mp.py:fact:31 "), compress(result.stdout)


@allpythons()
@mojo
def test_fork_full_metrics(py, mojo):
    result = austin("-i", "10ms", "-f", *python(py), target("target34.py"), mojo=mojo)
    assert py in (result.stderr or result.stdout), result.stderr or result.stdout

    assert len(processes(result.stdout)) == 1
    ts = threads(result.stdout)
    assert len(ts) == 2, ts

    assert has_pattern(result.stdout, "target34.py:keep_cpu_busy:32")
    assert not has_pattern(result.stdout, "Unwanted")

    meta = metadata(result.stdout)

    assert meta["mode"] == "full"

    wall, cpu, alloc, dealloc = sum_metrics(result.stdout)
    d = int(meta["duration"])

    assert 0 < 0.9 * d < wall < 2.1 * d
    assert 0 < cpu <= wall
    assert alloc * dealloc


@pytest.mark.parametrize("exposure", [1, 2])
@allpythons()
def test_fork_exposure(py, exposure):
    result = austin(
        "-i", "1ms", "-x", str(exposure), *python(py), target("sleepy.py"), "1"
    )
    assert result.returncode == 0, result.stderr or result.stdout

    assert has_pattern(result.stdout, "sleepy.py:<module>:"), compress(result.stdout)

    meta = metadata(result.stdout)

    assert meta["mode"] == "wall"

    d = int(meta["duration"])
    assert 900000 * exposure < d < 1100000 * exposure


@variants
@allpythons(min=(3, 11))
def test_qualnames(py, austin):
    result = austin("-i", "1ms", *python(py), target("qualnames.py"))
    assert py in (result.stderr or result.stdout), result.stderr or result.stdout

    assert len(processes(result.stdout)) == 1, compress(result.stdout)
    ts = threads(result.stdout)
    assert len(ts) == 1, compress(result.stdout)

    assert has_pattern(result.stdout, "qualnames.py:Foo.run"), compress(result.stdout)
    assert has_pattern(result.stdout, "qualnames.py:Bar.run"), compress(result.stdout)


@allpythons()
def test_no_logging(py, monkeypatch):
    monkeypatch.setenv("AUSTIN_NO_LOGGING", "1")
    result = austin("-i", "1ms", *python(py), target("target34.py"))
    assert has_pattern(result.stdout, "target34.py:keep_cpu_busy:3"), compress(
        result.stdout
    )
    assert result.returncode == 0, result.stderr or result.stdout



# This file is part of "austin" which is released under GPL.
#
# See file LICENCE or go to http://www.gnu.org/licenses/ for full license
# details.
#
# Austin is a Python frame stack sampler for CPython.
#
# Copyright (c) 2019 Gabriele N. Tornetta <phoenix1987@gmail.com>.
# All rights reserved.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import sys
import time


def cpu_bound():
    a = []
    for i in range(1000000):
        a.append(i)


if __name__ == "__main__":
    try:
        interval = float(sys.argv[1])
    except IndexError:
        interval = 0.7

    for n in range(2):
        cpu_bound()
        time.sleep(interval)



import os
import platform

PY3_EARLIEST = 8
PY3_LATEST = 13

try:
    REQUESTED_PYTHON_VERSIONS = [
        tuple(int(_) for _ in v.split("."))
        for v in os.getenv("AUSTIN_TESTS_PYTHON_VERSIONS", "").split(",")
    ]
except Exception:
    REQUESTED_PYTHON_VERSIONS = None


match platform.system():
    case "Darwin":
        PYTHON_VERSIONS = REQUESTED_PYTHON_VERSIONS or [
            (3, _) for _ in range(PY3_EARLIEST, PY3_LATEST + 1)
        ]
    case _:
        PYTHON_VERSIONS = REQUESTED_PYTHON_VERSIONS or [
            (3, _) for _ in range(PY3_EARLIEST, PY3_LATEST + 1)
        ]



import sys
from pathlib import Path
from test.cunit import SRC
from test.cunit import CModule


CFLAGS = ["-g", "-fprofile-arcs", "-ftest-coverage"]

sys.modules[__name__] = CModule.compile(SRC / Path(__file__).stem, cflags=CFLAGS)



import ctypes
import re
from ctypes import CDLL
from ctypes import POINTER
from ctypes import Structure
from ctypes import c_char_p
from ctypes import c_long
from ctypes import c_void_p
from ctypes import cast
from pathlib import Path
from subprocess import PIPE
from subprocess import STDOUT
from subprocess import run
from types import ModuleType
from typing import Any
from typing import Callable
from typing import Optional
from typing import Type

from pycparser import c_ast
from pycparser import c_parser
from pycparser.plyparser import ParseError


HERE = Path(__file__).resolve().parent
TEST = HERE.parent
ROOT = TEST.parent
SRC = ROOT / "src"


restrict_re = re.compile(r"__restrict \w+")

_header_head = r"""
#define __attribute__(x)
#define __extension__
#define __inline inline
#define __asm__(x)
#define __const const
#define __inline__ inline
#define __inline inline
#define __restrict
#define __signed__ signed
#define __GNUC_VA_LIST
#define __gnuc_va_list char
#define __thread
// typedef struct __builtin_va_list { } __builtin_va_list;
"""


def preprocess(source: Path) -> str:
    with source.open() as fin:
        code = _header_head + fin.read()
        return restrict_re.sub(
            "",
            run(
                ["gcc", "-E", "-P", "-"],
                stdout=PIPE,
                input=code.encode(),
                cwd=SRC,
            ).stdout.decode(),
        )


class CompilationError(Exception):
    pass


def compile(
    source: Path,
    cflags: list[str] = [],
    extra_sources: list[Path] = [],
    ldadd: list[str] = [],
    force: bool = False,
) -> None:
    binary = source.with_suffix(".so")

    if not force and binary.is_file():
        return

    result = run(
        [
            "gcc",
            "-shared",
            *cflags,
            "-o",
            str(binary),
            str(source),
            *(str(_) for _ in extra_sources),
            *ldadd,
        ],
        stdout=PIPE,
        stderr=STDOUT,
        cwd=SRC,
    )

    if result.returncode == 0:
        return

    raise CompilationError(result.stdout.decode())


C = CDLL("libc.so.6")


class CFunctionDef:
    def __init__(
        self, name: str, args: list[tuple[str, Type[ctypes._SimpleCData]]], rtype: Any
    ) -> None:
        self.name = name
        self.args = args
        self.rtype = rtype


class CTypeDef:
    def __init__(self, name: str, fields: list[str]) -> None:
        self.name = name
        self.fields = fields
        self.methods = []
        self.constructor = None


class CType(Structure):
    def __init__(self, *args: Any, **kwargs: Any) -> None:
        self.__cself__ = self.new(*args, **kwargs)

    def __del__(self) -> None:
        if len(self.destroy.__cmethod__.__args__) == 1:
            self.destroy()

    def __repr__(self) -> str:
        return f"<{self.__class__.__name__} CObject at {self.__cself__}>"


class CFunction:
    def __init__(self, cfuncdef: CFunctionDef, cfunc: Callable[..., Any]) -> None:
        self.__name__ = cfuncdef.name
        self.__args__ = [_[0] for _ in cfuncdef.args]
        self.__cfunc__ = cfunc

        # Prevent argument values from being truncated/mangled
        self.__cfunc__.argtypes = [_[1] for _ in cfuncdef.args]
        self.__cfunc__.restype = (
            cfuncdef.rtype if cfuncdef.rtype is not None else c_long
        )

        self._posonly = all(_ is None for _ in self.__args__)

    def check_args(self, args: tuple[Any], kwargs: dict[str, Any]) -> None:
        if self._posonly and kwargs:
            raise ValueError(f"{self} takes only positional arguments")

        nargs = len(args) + len(kwargs)
        if nargs != len(self.__args__):
            raise TypeError(
                f"{self} takes exactly {len(self.__args__)} arguments ({nargs} given)"
            )

    def __call__(self, *args: Any, **kwargs: Any) -> Any:
        self.check_args(args, kwargs)
        return self.__cfunc__(*args, **kwargs)

    def __repr__(self) -> None:
        return f"<CFunction '{self.__name__}'>"


class CMethod(CFunction):
    def __init__(
        self, cfuncdef: CFunctionDef, cfunc: Callable[..., Any], ctype: Any
    ) -> None:
        super().__init__(cfuncdef, cfunc)
        self.__ctype__ = ctype

    def __get__(self, obj: Any, objtype: Optional[Type] = None) -> None:
        def _(*args, **kwargs):
            cargs = [obj.__cself__, *args]
            self.check_args(cargs, kwargs)

            return self.__cfunc__(*cargs, **kwargs)

        _.__cmethod__ = self

        return _

    def __repr__(self) -> str:
        return f"<CMethod '{self.__name__}' of CType '{self.__ctype__.__name__}'>"


class CStaticMethod(CFunction):
    def __init__(
        self, cfuncdef: CFunctionDef, cfunc: Callable[..., Any], ctype: Any
    ) -> None:
        super().__init__(cfuncdef, cfunc)
        self.__ctype__ = ctype

    def __repr__(self) -> str:
        return f"<CStaticMethod '{self.__name__}' of CType '{self.__ctype__.__name__}'>"


class CMetaType(type(Structure)):
    def __new__(
        cls, cmodule: "CModule", ctypedef: CTypeDef, _: Optional[Any] = None
    ) -> Type["CMetaType"]:
        ctype = super().__new__(
            cls,
            ctypedef.name,
            (CType,),
            {"__cmodule__": cmodule},
        )

        constructor = getattr(cmodule.__binary__, f"{ctypedef.name[:-2]}_new")
        ctype.new = CStaticMethod(ctypedef.constructor, constructor, ctype)

        for method_def in ctypedef.methods:
            method_name = method_def.name
            method = getattr(cmodule.__binary__, f"{ctypedef.name[:-2]}__{method_name}")
            setattr(ctype, method_name, CMethod(method_def, method, ctype))

        ctype.__cname__ = ctypedef.name

        return ctype


class DeclCollector(c_ast.NodeVisitor):
    def __init__(self) -> None:
        self.types = {}
        self.functions = []

    def _get_type(self, node: c_ast.Node) -> None:
        return self.types[" ".join(node.type.type.names)]

    def visit_Typedef(self, node: c_ast.Node) -> None:
        if isinstance(node.type.type, c_ast.Struct) and node.type.declname.endswith(
            "_t"
        ):
            struct = node.type.type
            self.types[node.type.declname[:-2]] = CTypeDef(
                node.type.declname,
                [decl.name for decl in struct.decls],
            )

    def visit_Decl(self, node: c_ast.Node) -> None:
        if "extern" in node.storage:
            return

        if isinstance(node.type, c_ast.FuncDecl):
            func_name = node.name
            ret_type = node.type.type
            rtype = None
            if isinstance(ret_type, c_ast.PtrDecl):
                if "".join(ret_type.type.type.names) == "char":
                    rtype = c_char_p
                else:
                    rtype = c_void_p

            args = (
                [
                    (
                        (
                            _.name,
                            c_void_p if isinstance(_.type, c_ast.PtrDecl) else c_long,
                        )
                        if hasattr(_, "name")
                        else None
                    )
                    for _ in node.type.args.params
                ]
                if node.type.args is not None
                else []
            )
            if func_name.endswith("_new"):
                self.types[f"{func_name[:-4]}"].constructor = CFunctionDef(
                    "new", args, rtype
                )
            elif "__" in func_name:
                type_name, _, method_name = func_name.partition("__")
                if not type_name:
                    return
                self.types[type_name].methods.append(
                    CFunctionDef(method_name, args, rtype)
                )
            else:
                self.functions.append(CFunctionDef(func_name, args, rtype))

    def collect(self, decl: str) -> dict[str, CTypeDef]:
        parser = c_parser.CParser()
        try:
            ast = parser.parse(decl, filename="<preprocessed>")
        except ParseError as e:
            lines = decl.splitlines()
            line, col = (
                int(_) - 1 for _ in e.args[0].partition(" ")[0].split(":")[1:3]
            )
            for i in range(max(0, line - 4), min(line + 5, len(lines))):
                if i != line:
                    print(f"{i+1:5d}  {lines[i]}")
                else:
                    print(f"{i+1:5d}  \033[33;1m{lines[line]}\033[0m")
                    print(" " * (col + 5) + "\033[31;1m<<^\033[0m")
            raise

        self.visit(ast)
        return {
            k: v
            for k, v in self.types.items()
            if isinstance(v, CTypeDef) and v.constructor is not None
        }


class CModule(ModuleType):
    def __init__(self, source: Path) -> None:
        super().__init__(source.name, f"Generated from {source.with_suffix('.c')}")

        so_file = source.with_suffix(".so")
        while True:
            try:
                self.__binary__ = CDLL(so_file)
                break
            except OSError:
                # On parallel runs we might still be compiling the source file
                # so we try until we succeed.
                if not so_file.exists():
                    raise

        collector = DeclCollector()

        for name, ctypedef in collector.collect(
            preprocess(source.with_suffix(".h"))
        ).items():
            parts = name.split("_")
            py_name = "".join((_.capitalize() for _ in parts))
            setattr(self, py_name, CMetaType(self, ctypedef, None))

        for cfuncdef in collector.functions:
            name = cfuncdef.name
            try:
                cfunc = CFunction(cfuncdef, getattr(self.__binary__, name))
                setattr(self, name, cfunc)
            except AttributeError:
                # Not part of the binary
                pass

    def cglobal(self, name: str, ctype: str) -> Optional[Any]:
        cglobal = getattr(self.__binary__, name)
        if cglobal is None:
            return None

        c_type = getattr(ctypes, f"c_{ctype}")
        if c_type is None:
            return None

        return cast(cglobal, POINTER(c_type)).contents.value

    @classmethod
    def compile(
        cls,
        source: Path,
        cflags: list[str] = [],
        extra_sources: list[Path] = [],
        ldadd: list[str] = [],
    ):
        compile(source.with_suffix(".c"), cflags, extra_sources, ldadd)
        return cls(source)



# https://github.com/TheAlgorithms/Python/blob/98a4c2487814cdfe0822526e05c4e63ff6aef7d0/ciphers/rsa_key_generator.py


import random

from . import cryptomath_module as cryptoMath  # noqa: N812
from . import rabin_miller as rabinMiller  # noqa: N812


def main() -> None:
    print("Making key files...")
    make_key_files("rsa", 1024)
    print("Key files generation successful.")


def generate_key(key_size: int) -> tuple[tuple[int, int], tuple[int, int]]:
    print("Generating prime p...")
    p = rabinMiller.generate_large_prime(key_size)
    print("Generating prime q...")
    q = rabinMiller.generate_large_prime(key_size)
    n = p * q

    print("Generating e that is relatively prime to (p - 1) * (q - 1)...")
    while True:
        e = random.randrange(2 ** (key_size - 1), 2 ** (key_size))
        if cryptoMath.gcd(e, (p - 1) * (q - 1)) == 1:
            break

    print("Calculating d that is mod inverse of e...")
    d = cryptoMath.find_mod_inverse(e, (p - 1) * (q - 1))

    public_key = (n, e)
    private_key = (n, d)
    return (public_key, private_key)


def make_key_files(name: str, key_size: int) -> None:
    public_key, private_key = generate_key(key_size)
    print(f"\nWriting public key to file {name}_pubkey.txt...")
    with open(f"/tmp/{name}_pubkey.txt", "w") as out_file:
        out_file.write(f"{key_size},{public_key[0]},{public_key[1]}")

    print(f"Writing private key to file {name}_privkey.txt...")
    with open(f"/tmp/{name}_privkey.txt", "w") as out_file:
        out_file.write(f"{key_size},{private_key[0]},{private_key[1]}")


if __name__ == "__main__":
    main()



# This file is part of "austin" which is released under GPL.
#
# See file LICENCE or go to http://www.gnu.org/licenses/ for full license
# details.
#
# Austin is a Python frame stack sampler for CPython.
#
# Copyright (c) 2022 Gabriele N. Tornetta <phoenix1987@gmail.com>.
# All rights reserved.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from itertools import takewhile
from subprocess import check_output

from test.utils import allpythons
from test.utils import austin
from test.utils import compress
from test.utils import has_pattern
from test.utils import metadata
from test.utils import processes
from test.utils import python
from test.utils import samples
from test.utils import sum_metric
from test.utils import target
from test.utils import threads


@allpythons()
def test_pipe_wall_time(py):
    interval = 1
    result = austin("-Pi", f"{interval}ms", *python(py), target())
    assert result.returncode == 0

    meta = metadata(result.stdout)

    assert ".".join((str(_) for _ in meta["python"])).startswith(py), meta
    assert meta["mode"] == "wall", meta
    assert int(meta["duration"]) > 100000, meta
    assert meta["interval"] == str(interval * 1000), meta

    assert len(processes(result.stdout)) == 1
    ts = threads(result.stdout)
    assert len(ts) == 2, ts

    assert has_pattern(result.stdout, "target34.py:keep_cpu_busy:32")
    assert not has_pattern(result.stdout, "Unwanted")

    a = sum_metric(result.stdout)
    d = int(meta["duration"])

    assert 0 < 0.8 * d < a < 2.2 * d


@allpythons()
def test_pipe_cpu_time(py):
    result = austin("-sPi", "1ms", *python(py), target())
    assert result.returncode == 0

    meta = metadata(result.stdout)

    assert ".".join((str(_) for _ in meta["python"])).startswith(py), meta
    assert meta["mode"] == "cpu", meta
    assert int(meta["duration"]) > 100000, meta
    assert meta["interval"] == "1000", meta


@allpythons()
def test_pipe_wall_time_multiprocess(py):
    result = austin("-CPi", "1ms", *python(py), target())
    assert result.returncode == 0

    meta = metadata(result.stdout)

    assert meta["mode"] == "wall", meta
    assert int(meta["duration"]) > 100000, meta
    assert meta["interval"] == "1000", meta
    assert meta["multiprocess"] == "on", meta
    assert ".".join((str(_) for _ in meta["python"])).startswith(py), meta


@allpythons()
def test_pipe_wall_time_multiprocess_output(py, tmp_path):
    datafile = tmp_path / "test_pipe.austin"

    result = austin("-CPi", "1ms", "-o", str(datafile), *python(py), target())
    assert result.returncode == 0

    with datafile.open() as f:
        data = f.read()
        meta = metadata(data)

        assert meta, meta
        assert meta["mode"] == "wall", meta
        assert int(meta["duration"]) > 100000, meta
        assert meta["interval"] == "1000", meta
        assert meta["multiprocess"] == "on", meta
        assert ".".join((str(_) for _ in meta["python"])).startswith(py), meta

        assert has_pattern(data, "target34.py:keep_cpu_busy:32"), compress(data)

        a = sum(int(_.rpartition(" ")[-1]) for _ in samples(data))
        d = int(meta["duration"])

        assert 0 < 0.8 * d < a < 2.2 * d


@allpythons(min=(3, 11))
def test_python_version(py):
    result = austin("-Pi", "1s", *python(py), target())
    assert result.returncode == 0

    meta = metadata(result.stdout)

    reported_version = ".".join((str(_) for _ in meta["python"]))

    # Take the version from the interpreter itself, discarding anything after
    # the patchlevel.
    actual_version = "".join(
        takewhile(
            lambda _: _.isdigit() or _ == ".",
            (check_output([*python(py), "-V"]).decode().strip().partition(" ")[-1]),
        )
    )

    assert reported_version == actual_version, meta



# This file is part of "austin" which is released under GPL.
#
# See file LICENCE or go to http://www.gnu.org/licenses/ for full license
# details.
#
# Austin is a Python frame stack sampler for CPython.
#
# Copyright (c) 2023 Gabriele N. Tornetta <phoenix1987@gmail.com>.
# All rights reserved.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from pathlib import Path
from test.utils import allpythons, austin, python, target

from austin.format.mojo import MojoFile, MojoFrame


@allpythons(min=(3, 11))
def test_mojo_column_data(py, tmp_path: Path):
    datafile = tmp_path / "test_mojo_column.austin"

    result = austin(
        "-i", "100", "-o", str(datafile), *python(py), target("column.py"), mojo=True
    )
    assert result.returncode == 0, result.stderr or result.stdout

    def strip(f):
        return (f.scope.string.value, f.line, f.line_end, f.column, f.column_end)

    with datafile.open("rb") as f:
        frames = {
            strip(_)
            for _ in MojoFile(f).parse()
            if isinstance(_, MojoFrame)
            and _.filename.string.value.endswith("column.py")
        }

        assert frames & {
            ("<module>", 16, 19, 5, 2),
            ("<listcomp>", 16, 19, 5, 2),
            ("lazy", 6, 6, 9, 19),
            ("<listcomp>", 17, 17, 5, 17),
        }


@allpythons(max=(3, 10))
def test_mojo_no_column_data(py, tmp_path: Path):
    """
    Test that no other location information is present apart from the line
    number for Python versions prior to 3.11.
    """
    datafile = tmp_path / "test_mojo_column.austin"

    result = austin(
        "-i", "100", "-o", str(datafile), *python(py), target("column.py"), mojo=True
    )
    assert result.returncode == 0, result.stderr or result.stdout

    def strip(f):
        return

    with datafile.open("rb") as f:
        assert {
            (e.line_end, e.column, e.column_end)
            for e in MojoFile(f).parse()
            if isinstance(e, MojoFrame)
        } == {(0, 0, 0)}



# This file is part of "austin" which is released under GPL.
#
# See file LICENCE or go to http://www.gnu.org/licenses/ for full license
# details.
#
# Austin is a Python frame stack sampler for CPython.
#
# Copyright (c) 2022 Gabriele N. Tornetta <phoenix1987@gmail.com>.
# All rights reserved.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import platform
import sys
from test.utils import austin
from test.utils import no_sudo
from test.utils import run_python
from test.utils import target

import pytest


def test_cli_no_arguments():
    result = austin()
    assert result.returncode == 0
    assert "Usage:" in result.stdout
    assert not result.stderr


def test_cli_no_python():
    match platform.system():
        case "Windows":
            cmd = ["cmd.exe", "/c", "sleep 2"]
        case _:
            cmd = ["bash", "-c", "sleep 2"]

    result = austin(*cmd, expect_fail=32)

    assert result.returncode == 32
    assert "not a Python" in result.stderr or "Cannot launch" in result.stderr


def test_cli_short_lived():
    result = austin(
        "src\\austin.exe" if platform.system() == "Windows" else "src/austin",
        sys.executable,
        "-c",
        "print('Hello World')",
        expect_fail=33,
    )
    assert result.returncode == 33
    assert "too quickly" in result.stderr


def test_cli_invalid_command():
    result = austin("snafubar", expect_fail=33)
    assert result.returncode == 33
    assert "Cannot launch" in (result.stderr or result.stdout)


def test_cli_invalid_pid():
    result = austin("-p", "9999999", expect_fail=36)

    assert result.returncode == 36
    assert "Cannot attach" in result.stderr


@pytest.mark.skipif(
    platform.system() == "Windows", reason="No permission issues on Windows"
)
@no_sudo
def test_cli_permissions():
    with run_python("3", target("sleepy.py")) as p:
        result = austin("-i", "1ms", "-p", str(p.pid), expect_fail=37)
        assert result.returncode == 37, result.stderr
        assert "Insufficient permissions" in result.stderr, result.stderr


@pytest.mark.skipif(
    platform.system() != "Darwin",
    reason="Only Darwin requires sudo in all cases",
)
@no_sudo
def test_cli_permissions_darwin():
    result = austin(
        "-i",
        "1ms",
        "python3.10",
        "-c",
        "from time import sleep; sleep(1)",
        expect_fail=37,
    )
    assert result.returncode == 37, result.stderr
    assert "Insufficient permissions" in result.stderr, result.stderr



# This file is part of "austin" which is released under GPL.
#
# See file LICENCE or go to http://www.gnu.org/licenses/ for full license
# details.
#
# Austin is a Python frame stack sampler for CPython.
#
# Copyright (c) 2019 Gabriele N. Tornetta <phoenix1987@gmail.com>.
# All rights reserved.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

# source: https://lobste.rs/s/qairy5/austin_python_frame_stack_sampler_for

import multiprocessing


def fact(n):
    f = 1
    for i in range(1, n + 1):
        f *= i
    return f


def do(N):
    n = 1
    for _ in range(N):
        fact(n)
        n += 1


if __name__ == "__main__":
    import sys

    try:
        nproc = int(sys.argv[1])
    except Exception:
        nproc = 2

    processes = []
    for _ in range(nproc):
        process = multiprocessing.Process(target=do, args=(3000,))
        process.start()
        processes.append(process)

    for process in processes:
        process.join(timeout=5)



from pathlib import Path

HERE = Path(__file__).parent


def get_current_version_from_changelog() -> str:
    with (HERE.parent / "ChangeLog").open() as cl:
        for line in cl.readlines():
            if not line or line.startswith(" "):
                continue
            _, _, version = line.partition(" ")
            if version[0] == "v":
                return version[1:].strip()



from ctypes import c_void_p
from test.cunit import C
from test.cunit.cache import Chain
from test.cunit.cache import HashTable
from test.cunit.cache import Lookup
from test.cunit.cache import LruCache
from test.cunit.cache import Queue
from test.cunit.cache import QueueItem

import pytest


NULL = 0
C.free.argtypes = [c_void_p]
C.malloc.restype = c_void_p


def test_queue_item():
    value = C.malloc(16)
    queue_item = QueueItem(value, 42)
    assert queue_item.__cself__

    queue_item.destroy(C.free)


@pytest.mark.parametrize("qsize", [0, 10, 100, 1000])
def test_queue(qsize):
    q = Queue(qsize, C.free)

    assert q.is_empty()
    assert qsize == 0 or not q.is_full()

    assert q.dequeue() is NULL

    values = [C.malloc(16) for _ in range(qsize)]
    assert all(values)

    for k, v in enumerate(values):
        assert q.enqueue(v, k)

    assert qsize == 0 or not q.is_empty()
    assert q.is_full()
    assert q.enqueue(42, 42) is None

    assert values == [q.dequeue() for _ in range(qsize)]


def test_chain():
    head = Chain(0, NULL)

    head.add(42, 24)

    assert head.has(42)
    assert not head.has(43)
    assert head.find(42) == 24
    assert head.find(43) is NULL

    head.remove(42)
    assert not head.has(42)
    assert head.find(42) is NULL


def test_hash_table():
    t = HashTable(10)
    assert t.get(42) is NULL

    t.set(42, 24)
    assert t.get(42) == 24

    getattr(t, "del")(42)
    assert t.get(42) is NULL


def test_hash_table_full():
    t = HashTable(10)

    for i in range(11):
        t.set(42 + i, i + 1)

    assert t.get(52) is NULL

    assert t.get(42) == 1

    t.set(42, 24)
    assert t.get(42) == 24


def test_lru_cache():
    c = LruCache(10, C.free)

    assert not c.is_full()

    assert c.maybe_hit(42) is NULL

    values = [(42 + i, C.malloc(8)) for i in range(10)]
    for k, v in values:
        c.store(k, v)

    assert c.is_full()

    for i in range(8):
        assert c.maybe_hit(42 + i) == values[i][1]

    assert c.maybe_hit(51) == values[9][1]

    c.store(100, C.malloc(8))
    assert c.maybe_hit(50) is NULL


def test_lru_cache_expand():
    c = LruCache(0, C.free)

    # Fill the cache to the initial size
    values = [(i + 1, C.malloc(8)) for i in range(1024)]
    for k, v in values:
        c.store(k, v)

    # Check that we can retrieve the values
    for k, v in values:
        assert c.maybe_hit(k) == v

    # Check that the cache is now full
    assert c.is_full()

    # Add a new value
    new_value = C.malloc(8)
    c.store(1025, new_value)

    # Check that the cache has been resized and not full
    assert not c.is_full()

    values.append((1025, new_value))

    # Check that we still have all the items
    for k, v in values:
        assert c.maybe_hit(k) == v


def test_lookup():
    lu = Lookup(8)

    assert not lu.get(42)

    for i in range(1000):
        lu.set(42 + i, i + 1)

    for i in range(1000):
        assert lu.get(42 + i) == i + 1

    getattr(lu, "del")(42)
    assert not lu.get(42)

    lu.clear()

    for i in range(1000):
        assert not lu.get(42 + i)



from time import sleep


def lazy(n):
    for _ in range(n):
        sleep(0.1)
        yield _


def fib(n):
    a, b = 0, 1
    for _ in range(n):
        yield a
        a, b = b, a + b

a = [
    list(fib(_))
    for _ in lazy(30)
]

print(a)



#!/usr/bin/env python3

# This file is part of "austin" which is released under GPL.
#
# See file LICENCE or go to http://www.gnu.org/licenses/ for full license
# details.
#
# Austin is a Python frame stack sampler for CPython.
#
# Copyright (c) 2019 Gabriele N. Tornetta <phoenix1987@gmail.com>.
# All rights reserved.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import threading


def keep_cpu_busy():
    a = []
    for i in range(10_000_000):
        a.append(i)
        if i % 1000000 == 0:
            print("Unwanted output " + str(i))


if __name__ == "__main__":
    threading.Thread(target=keep_cpu_busy).start()
    keep_cpu_busy()



import sys
from itertools import product
from pathlib import Path

sys.path.insert(0, str(Path(__file__).parent.parent))

import tarfile
from io import BytesIO
from test.utils import Variant
from urllib.error import HTTPError
from urllib.request import urlopen
import json


def get_latest_release() -> str:
    with urlopen(
        "https://api.github.com/repos/p403n1x87/austin/releases/latest"
    ) as stream:
        return json.loads(stream.read().decode("utf-8"))["tag_name"].strip("v")


def download_release(version: str, dest: Path, variant_name: str = "austin") -> Variant:
    if version == "dev":
        return Variant(f"src/{variant_name}")

    binary_dest = dest / version
    binary = binary_dest / variant_name

    if not binary.exists():
        prefix = "https://github.com/p403n1x87/austin/releases/download/"
        for flavour, v in product({"-gnu", ""}, {"", "v"}):
            try:
                with urlopen(
                    f"{prefix}v{version}/{variant_name}-{v}{version}{flavour}-linux-amd64.tar.xz"
                ) as stream:
                    buffer = BytesIO(stream.read())
                    binary_dest.mkdir(parents=True, exist_ok=True)
                    tar = tarfile.open(fileobj=buffer, mode="r:xz")
                    tar.extract(variant_name, str(binary_dest))
            except HTTPError:
                continue
            break
        else:
            raise RuntimeError(f"Could not download Austin version {version}")

    variant = Variant(str(binary))

    out = variant("-V").stdout
    assert f"{variant_name} {version}" in out, (f"{variant_name} {version}", out)

    return variant


def download_latest(dest: Path, variant_name: str = "austin") -> Variant:
    return download_release(get_latest_release(), dest, variant_name)


def get_dev(variant_name: str = "austin") -> Variant:
    return download_release("dev", None, variant_name)



# https://github.com/TheAlgorithms/Python/blob/98a4c2487814cdfe0822526e05c4e63ff6aef7d0/ciphers/cryptomath_module.py


def gcd(a: int, b: int) -> int:
    while a != 0:
        a, b = b % a, a
    return b


def find_mod_inverse(a: int, m: int) -> int:
    if gcd(a, m) != 1:
        raise ValueError(f"mod inverse of {a!r} and {m!r} does not exist")
    u1, u2, u3 = 1, 0, a
    v1, v2, v3 = 0, 1, m
    while v3 != 0:
        q = u3 // v3
        v1, v2, v3, u1, u2, u3 = (u1 - q * v1), (u2 - q * v2), (u3 - q * v3), v1, v2, v3
    return u1 % m



# Run as python3 scripts/validation.py from the repository root directory.
# Ensure dependencies from requirements-val.txt are installed.

from argparse import ArgumentParser
from collections import Counter, namedtuple
from io import BytesIO
from itertools import chain
import os
from pathlib import Path
import re
import sys
import typing as t
import codecs

import common

import numpy as np
from scipy.stats import f

from austin.format.mojo import (
    MojoFile,
    MojoStack,
    MojoFrameReference,
    MojoMetric,
    MojoFrame,
)
from test.utils import python, target

Scenario = namedtuple("Scenario", ["title", "variant", "args"])

PYTHON = (
    python(os.getenv("AUSTIN_TESTS_PYTHON_VERSIONS"))
    if "AUSTIN_TESTS_PYTHON_VERSIONS" in os.environ
    else [sys.executable]
)

SCENARIOS = [
    Scenario(
        "Wall time",
        "austin",
        (
            "-i",
            "500",
            *PYTHON,
            target("target34.py"),
        ),
    ),
    Scenario(
        "CPU time",
        "austin",
        (
            "-si",
            "500",
            *PYTHON,
            target("target34.py"),
        ),
    ),
    Scenario(
        "Wall time [multiprocessing]",
        "austin",
        (
            "-Ci",
            "500",
            *PYTHON,
            target("target_mp.py"),
        ),
    ),
    Scenario(
        "CPU time [multiprocessing]",
        "austin",
        (
            "-Csi",
            "500",
            *PYTHON,
            target("target_mp.py"),
        ),
    ),
]


class AustinFlameGraph(dict):
    def __call__(self, x):
        return self.get(x, 0)

    def __add__(self, other):
        m = self.__class__(self)
        for k, v in other.items():
            n = m.setdefault(k, v.__class__()) + v
            if not n and k in m:
                del m[k]
                continue
            m[k] = n
        return m

    def __mul__(self, other):
        m = self.__class__(self)
        for k, v in self.items():
            n = v * other
            if not n and k in m:
                del m[k]
                continue
            m[k] = n
        return m

    def __rmul__(self, other):
        return self.__mul__(other)

    def __truediv__(self, other):
        return self * (1 / other)

    def __rtruediv__(self, other):
        return self.__div__(other)

    def __sub__(self, other):
        return self + (-other)

    def __neg__(self):
        m = self.__class__(self)
        for k, v in m.items():
            m[k] = -v
        return m

    def supp(self):
        return set(self.keys())

    def to_list(self, domain: list) -> list:
        return [self(v) for v in domain]

    @classmethod
    def from_list(cls, stacks: t.List[t.Tuple[str, int]]) -> "AustinFlameGraph":
        return sum((cls({stack: metric}) for stack, metric in stacks), cls())

    @classmethod
    def from_mojo(cls, data: bytes) -> "AustinFlameGraph":
        fg = cls()

        stack: t.List[str] = []
        metric = 0

        def serialize(frame: MojoFrame) -> str:
            return ":".join(
                (
                    frame.filename.string.value,
                    frame.scope.string.value,
                    str(frame.line),
                    str(frame.line_end),
                    str(frame.column),
                    str(frame.column_end),
                )
            )

        for e in MojoFile(BytesIO(data)).parse():
            if isinstance(e, MojoStack):
                if stack:
                    fg += cls({";".join(stack): metric})
                stack.clear()
                metric = 0
            elif isinstance(e, MojoFrameReference):
                stack.append(serialize(e.frame))
            elif isinstance(e, MojoMetric):
                metric = e.value

        return fg


def hotelling_two_sample_test(X, Y) -> float:
    nx, p = X.shape
    ny, q = Y.shape

    assert p == q, "X and Y must have the same dimensionality"

    dof = nx + ny - p - 1

    assert (
        dof > 0
    ), f"X ({nx}x{p}) and Y ({ny}x{q}) must have at least p ({p}) + 1 samples"

    g = dof / p / (nx + ny - 2) * (nx * ny) / (nx + ny)

    x_mean = np.mean(X, axis=0)
    y_mean = np.mean(Y, axis=0)
    delta = x_mean - y_mean

    x_cov = np.cov(X, rowvar=False)
    y_cov = np.cov(Y, rowvar=False)
    pooled_cov = ((nx - 1) * x_cov + (ny - 1) * y_cov) / (nx + ny - 2)

    # Compute the F statistic from the Hotelling T^2 statistic
    statistic = g * delta.transpose() @ np.linalg.inv(pooled_cov) @ delta
    f_pdf = f(p, dof)

    return 1 - f_pdf.cdf(statistic)


def compare(
    x: t.List[AustinFlameGraph],
    y: t.List[AustinFlameGraph],
    threshold: t.Optional[float] = None,
) -> float:
    domain = list(set().union(*(_.supp() for _ in chain(x, y))))

    if threshold is not None:
        c = Counter()
        for _ in chain(x, y):
            c.update(_.supp())
        domain = sorted([k for k, v in c.items() if v >= threshold])

    X = np.array([f.to_list(domain) for f in x], dtype=np.int32)
    Y = np.array([f.to_list(domain) for f in y], dtype=np.int32)

    return hotelling_two_sample_test(X, Y)


def validate(args, variant: str = "austin", runs: int = 10) -> float:
    austin_latest = common.download_latest(dest=Path("/tmp"), variant_name=variant)
    austin_dev = common.get_dev(variant_name=variant)

    return compare(
        *(
            [
                AustinFlameGraph.from_mojo(
                    austin(
                        *args,
                        mojo=True,
                        convert=False,
                    ).stdout
                )
                for _ in range(runs)
            ]
            for austin in (austin_latest, austin_dev)
        ),
        threshold=runs,  # Keep only the stacks that are present in all runs
    )


if __name__ == "__main__":
    argp = ArgumentParser()

    argp.add_argument(
        "-k",
        type=re.compile,
        help="Run data validation scenarios that match the given regular expression",
    )

    argp.add_argument(
        "-n",
        type=int,
        default=30,
        help="Number of profiles to collect",
    )

    argp.add_argument(
        "-i",
        "--ignore-errors",
        action="store_true",
        help="Ignore encoding errors",
    )

    argp.add_argument(
        "-p",
        "--p-value",
        type=float,
        default=0.01,
        help="p-value threshold",
    )

    opts = argp.parse_args()

    if opts.ignore_errors:
        codecs.register_error("strict", codecs.ignore_errors)

    print("# Austin Data Validation\n")

    failures: t.List[Scenario] = []
    for scenario in SCENARIOS:
        print(
            f"Validating {scenario.title} ...                                ",
            end="\r",
            flush=True,
            file=sys.stderr,
        )
        if (p := validate(scenario.args, scenario.variant, runs=opts.n)) < opts.p_value:
            failures.append((scenario, p))

    if failures:
        print("💥 The following scenarios failed to validate:\n")

        for scenario, p in failures:
            print(f"- {scenario.title} [{p:.2%}]")

        exit(1)

    print(f"✨ 🍰 ✨ All {len(SCENARIOS)} scenarios validated successfully!")



import sys
from pathlib import Path
from test.cunit import SRC
from test.cunit import CModule


CFLAGS = ["-g", "-fprofile-arcs", "-ftest-coverage", "-fPIC"]

EXTRA_SOURCES = [
    SRC / "argparse.c",
    SRC / "logging.c",
]

sys.modules[__name__] = CModule.compile(
    SRC / Path(__file__).stem, cflags=CFLAGS, extra_sources=EXTRA_SOURCES
)
